# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# cSpell: ignore sdtype, sdio

import gc
import os
from pathlib import Path
from typing import Union, Optional
import xml.etree.ElementTree as ET

from ansys.scadeone.core.common.exception import ScadeOneException
import ansys.scadeone.core.svc.simdata as sd

from .cmd_context import SDContext
from .swan_context import SwanContext
from .utils import ConverterLogger, SDType


class STPRecord:
    """STP Procedure record"""

    sss_file_extension = ".sss"

    def __init__(self, name, inits=None, preambles=None, scenarios=None) -> None:
        self._name = name
        self._inits: list[Path] = inits if inits else []
        self._preambles: list[Path] = preambles if preambles else []
        self._scenarios: list[Path] = scenarios if scenarios else []

    @property
    def name(self) -> str:
        """Gets record name"""
        return self._name

    @property
    def inits(self) -> list[Path]:
        """Gets record inits"""
        return self._inits

    @property
    def preambles(self) -> list[Path]:
        """Gets record preambles"""
        return self._preambles

    @property
    def scenarios(self) -> list[Path]:
        """Gets record scenarios"""
        return self._scenarios

    @inits.setter
    def inits(self, inits: list[Path]) -> None:
        """Sets init files"""
        self._inits = inits

    @preambles.setter
    def preambles(self, preambles: list[Path]) -> None:
        """Sets preamble files"""
        self._preambles = preambles

    @scenarios.setter
    def scenarios(self, scenarios: list[Path]) -> None:
        """Sets scenario files"""
        self._scenarios = scenarios

    def start_convert(
        self, swan_ctx: SwanContext, out_dir: Path, gc_disable: bool = False
    ) -> list[str]:
        """Converts all the sss files for the record"""
        ConverterLogger.getLogger(self._name)
        ConverterLogger.info("Starting Conversion of record: " + self._name)

        if not out_dir.exists():
            os.makedirs(out_dir)

        ctx = SDContext(swan_ctx)
        for init in self._inits:
            if init.suffix == STPRecord.sss_file_extension:
                ctx.parse_file(init)
            else:
                ConverterLogger.warning(f"Unsupported init file was not converted: {init.name}")

        for preamble in self._preambles:
            if preamble.suffix == STPRecord.sss_file_extension:
                ctx.parse_file(preamble)
            else:
                ConverterLogger.warning(
                    f"Unsupported preamble file was not converted: {preamble.name}"
                )

        for scenario in self._scenarios:
            if scenario.suffix == STPRecord.sss_file_extension:
                ctx.parse_file(scenario)
            else:
                ConverterLogger.warning(
                    f"Unsupported scenario file was not converted: {scenario.name}"
                )

        sd_name = out_dir / (self._name + ".sd")
        sd_name_checks = out_dir / (self._name + "_checks.sd")

        if gc_disable:
            gc.disable()

        ConverterLogger.getLogger(self._name + " (inputs)")

        sd_file = sd.create_file(str(sd_name))
        ctx.sdf = sd_file
        ctx.sdio_type = SDType.INPUT
        self._write_file(sd_name, ctx)

        ConverterLogger.getLogger(self._name + " (checks)")

        sd_file = sd.create_file(str(sd_name_checks))
        ctx.sdf = sd_file
        ctx.sdio_type = SDType.OUTPUT
        self._write_file(sd_name_checks, ctx)

        if gc_disable:
            gc.enable()

        return [sd_name, sd_name_checks]

    def _write_file(self, sd_name: Path, ctx: "SDContext"):
        """Writes the sd file

        Parameters
        ----------
        sd_name : str
            name of created file
        ctx : SDContext
            simdata context of writing file
        """
        ctx.eval_cmds()
        ctx.sdf.close()
        sd_fd = sd.open_file(str(sd_name))
        sd_content = str(sd_fd)
        sd_fd.close()
        if ConverterLogger.SD_VERBOSE:
            print(f"Created {sd_name}")
            print(sd_content)
        ConverterLogger.debug(f"Created {sd_name}: {sd_content}")

    def associate_sd_files(self):
        """Associates the new sd files to project"""
        raise NotImplementedError

    def get_sss_files_names(self) -> str:
        """Gets all file names of all sss file paths"""
        return (
            f"inits {[path.name for path in self._inits]}"
            f"\npreambles {[path.name for path in self._preambles]}"
            f"\nscenarios {[path.name for path in self._scenarios]}"
        )

    def __str__(self):
        return f"inits {self._inits}\npreambles {self._preambles}\nscenarios {self._scenarios}"


class STPContext:
    """STP data that includes all paths to SSS files"""

    def __init__(self, swan_ctx: SwanContext, root_operator: Optional[str] = None) -> None:
        self._output_dir: Path = Path(".")
        self._stp_path: Path = None
        self._records: list[STPRecord] = []
        self._generated_files = []
        self._stp_root = None
        self._swan_ctx = swan_ctx
        self._operator = None
        self._root_operator = root_operator

    @property
    def swan_context(self) -> SwanContext:
        return self._swan_ctx

    @property
    def stp_path(self) -> Path:
        """Returns STP file path"""
        return self._stp_path

    @stp_path.setter
    def stp_path(self, path: Union[str, Path]):
        """Sets STP file path"""
        if isinstance(path, str):
            path = Path(path)
        self._stp_path = path

    @property
    def proj_path(self) -> str:
        """Returns swan project path"""
        return self._proj_path

    @proj_path.setter
    def proj_path(self, proj_path: str) -> None:
        """Sets swan project path, required to get variable types"""
        self._proj_path = proj_path

    def _load_stp(self):
        if self._stp_path is None:
            ConverterLogger.exception("No STP path was defined for this context")
        try:
            tree = ET.parse(self._stp_path)
            root = tree.getroot()
            self._output_dir = self._output_dir / root.attrib["name"]
            # use user provided operator if it exists, or get from STP file
            if self._root_operator is not None:
                self._operator = self._root_operator
            else:
                if "operator" not in root.attrib:
                    raise Exception("Missing operator attribute")
                self._operator = root.attrib["operator"]
            self._stp_root = root

        except Exception as e:
            ConverterLogger.exception(f"Error while loading STP file : {e}")

    def load_record(self, record_name: str):
        """Reads STP file and loads all sss file paths for the specified record name"""

        self._load_stp()
        try:
            for record in self._stp_root.findall(".//Record"):
                if record.attrib["name"] == record_name:
                    self._parse_xml_record(record)
                    return
        except Exception as e:
            ConverterLogger.exception(f"Error while loading STP file : {e}")
        ConverterLogger.exception(f"No record was found for {record_name} in STP file")

    def load_all_records(self) -> None:
        """Reads STP file and appends sss paths from EACH of the STP records"""

        self._load_stp()
        try:
            for record in self._stp_root.findall(".//Record"):
                self._parse_xml_record(record)
        except Exception as e:
            ConverterLogger.exception(f"Error while loading STP file : {e}")

    def _parse_xml_record(self, record: ET.Element) -> None:
        """Parses a record element in XML"""

        stp_record = STPRecord(record.attrib["name"])
        record_inits = record.find("inits")
        record_preambles = record.find("preambles")
        record_scenarios = record.find("scenarios")
        if record_inits is not None:
            inits = []
            for init in record_inits:
                sss_path = self._stp_path.parent / init.attrib["persistAs"]
                inits.append(sss_path.resolve())
            stp_record.inits = inits
        if record_preambles is not None:
            preambles = []
            for preamble in record_preambles:
                sss_path = self._stp_path.parent / preamble.attrib["persistAs"]
                preambles.append(sss_path.resolve())
            stp_record.preambles = preambles
        if record_scenarios is not None:
            scenarios = []
            for scenario in record_scenarios:
                sss_path = self._stp_path.parent / scenario.attrib["persistAs"]
                scenarios.append(sss_path.resolve())
            stp_record.scenarios = scenarios
        self._records.append(stp_record)

    @property
    def generated_files(self) -> list[str]:
        """Gets the list of created files"""
        return self._generated_files

    def get_record_names(self) -> list[str]:
        """Gets all record names in a list"""
        return [record.name for record in self._records]

    @property
    def output_dir(self) -> Path:
        """Gets output directory for simdata files"""
        return self._output_dir

    @output_dir.setter
    def output_dir(self, path: Union[Path, str]) -> None:
        """Sets output directory for simdata files"""
        self._output_dir = Path(path)

    def start_all_converts(self, gc_disable: bool = False) -> bool:
        """Converts all the loaded sss files"""
        self._swan_ctx.get_operator(self._operator)
        try:
            for record in self._records:
                self._generated_files += record.start_convert(
                    self.swan_context, self._output_dir, gc_disable
                )
            return True
        except ScadeOneException as e:
            ConverterLogger.error(f"Error has occurred while converting files: {e}")
            return False
        except Exception as e:
            ConverterLogger.error(f"Internal error has occurred while converting files: {e}")
            print(f"Internal error {e}")
            return False

    def associate_sd_files():
        """Associates the new sd files to project"""
        raise NotImplementedError

    def __str__(self):
        return (
            f"STP path: {str(self._stp_path)}"
            f"\nSwan path: {str(self._proj_path)}"
            f"\nRecords: {[str(record.name) for record in self._records]}"
        )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# cSpell:ignore prio
# pylint: disable=too-many-arguments

from abc import ABC
from typing import List, Optional, Union

import ansys.scadeone.core.swan.common as common
import ansys.scadeone.core.swan.scopes as scopes

from .expressions import Literal, Pattern


class LHSItem(common.SwanItem):  # numpydoc ignore=PR01
    """Defines an item on the left-hand side of an equation, an ID, or underscore '_'.

    Parameters
    ----------
    id : Identifier (optional)
        Identifier or None for underscore value.
    """

    def __init__(self, id: Union[common.Identifier, str] = None) -> None:
        super().__init__()
        self._id = id

    @property
    def id(self) -> Union[common.Identifier, None]:
        """Returns id value or None."""
        return self._id

    @property
    def is_underscore(self) -> bool:
        """True when LHSItem is '_'."""
        return self._id is None


class EquationLHS(common.SwanItem):  # numpydoc ignore=PR01
    """Equation left-hand side part:

    *lhs* ::= ( ) | *lhs_item* {{ , *lhs_item* }} [[ , .. ]]

    """

    def __init__(self, lhs_items: List[LHSItem], is_partial_lhs: Optional[bool] = False) -> None:
        super().__init__()
        self._lhs_items = lhs_items
        self._is_partial_lhs = is_partial_lhs

    @property
    def is_partial_lhs(self) -> bool:
        """True when lhs list is partial (syntax: final '..' not in the list."""
        return self._is_partial_lhs

    @property
    def lhs_items(self) -> List[LHSItem]:
        """Return left-hand side list."""
        return self._lhs_items


class ExprEquation(common.Equation):  # numpydoc ignore=PR01
    """Flows definition using an expression:

    *equation* ::= *lhs* [luid] = *expr*"""

    def __init__(
        self,
        lhs: EquationLHS,
        expr: common.Expression,
        luid: Optional[common.Luid] = None,
    ) -> None:
        super().__init__()
        self._lhs = lhs
        self._expr = expr
        self._luid = luid
        common.SwanItem.set_owner(self, expr)

    @property
    def lhs(self) -> EquationLHS:
        """Left-hand side of the equation."""
        return self._lhs

    @property
    def expr(self) -> common.Expression:
        """Equation expression."""
        return self._expr

    @property
    def luid(self) -> Union[common.Luid, None]:
        """Equation LUID."""
        return self._luid


# Definition by cases: state machines and activate if/when
# ========================================================
class DefByCase(common.Equation, ABC):  # numpydoc ignore=PR01
    """Base class for state machine and active if/when equations."""

    def __init__(
        self,
        lhs: Optional[EquationLHS] = None,
        name: Optional[common.Luid] = None,
        is_equation: bool = False,
    ) -> None:
        common.Equation.__init__(self)
        self._lhs = lhs
        self._name = name
        self._is_equation = is_equation

    @property
    def lhs(self) -> Union[EquationLHS, None]:
        """Left-hand side of the equation, may be None."""
        return self._lhs

    @property
    def name(self) -> Union[common.Luid, None]:
        """Return name or None if no name."""
        return self._name

    @property
    def is_equation(self) -> bool:
        """True when the object is an equation."""
        return self._is_equation


# State Machines
# ============================================================


class StateMachineItem(common.SwanItem, ABC):  # numpydoc ignore=PR01
    """Base class for state machine items (states and transitions)."""

    def __init__(self) -> None:
        common.SwanItem.__init__(self)


class StateRef(common.SwanItem):  # numpydoc ignore=PR01
    """State identification:

    *state_ref* ::= ID | LUNUM

    The class is also used for transition declaration or target
    (**restart**/**resume**) where one has either an ID or a LUNUM.
    """

    def __init__(
        self,
        lunum: Optional[common.Lunum] = None,
        id: Optional[common.Identifier] = None,
    ) -> None:
        super().__init__()
        self._lunum = lunum
        self._id = id

    @property
    def lunum(self) -> Union[common.Luid, None]:
        """Lunum part, possible None."""
        return self._lunum

    @property
    def id(self) -> Union[common.Identifier, None]:
        """Id part, possible None."""
        return self._id


class Fork(common.SwanItem):  # numpydoc ignore=PR01
    """Base class for fork-related classes."""

    def __init__(self) -> None:
        super().__init__()


class Target(common.SwanItem):  # numpydoc ignore=PR01
    """Arrow target as a state reference and kind."""

    def __init__(self, target: StateRef, is_resume: Optional[bool] = False) -> None:
        super().__init__()
        self._is_resume = is_resume
        self._target = target

    @property
    def is_resume(self) -> bool:
        """True when is **resume**, else **restart**."""
        return self._is_resume

    @property
    def is_restart(self) -> bool:
        """True when is **restart**, else **resume**."""
        return not self.is_resume

    @property
    def target(self) -> StateRef:
        """Target reference"""
        return self._target


class Arrow(common.SwanItem):  # numpydoc ignore=PR01
    """Encode an arrow, with or without guard:

    | *guarded_arrow* ::= ( *expr* ) *arrow*
    | *arrow* ::= [[ *scope* ]] (( *target* | *fork* ))
    """

    def __init__(
        self,
        guard: Union[common.Expression, None],
        action: Union[scopes.Scope, None],
        target: Optional[Target] = None,
        fork: Optional[Fork] = None,
    ) -> None:
        super().__init__()
        self._guard = guard
        self._action = action
        self._target = target
        self._fork = fork

    @property
    def guard(self) -> Union[scopes.Scope, None]:
        """Arrow guard or None."""
        return self._guard

    @property
    def action(self) -> Union[scopes.Scope, None]:
        """Arrow action or None."""
        return self._action

    @property
    def target(self) -> Union[Target, None]:
        """Arrow target."""
        return self._target

    @property
    def fork(self) -> Union[Fork, None]:
        """Arrow fork."""
        return self._fork

    @property
    def is_valid(self) -> bool:
        "Check whether the arrow has a target or a fork."
        return bool(self.target) != bool(self.fork)


class ForkTree(Fork):  # numpydoc ignore=PR01
    """Fork as a tree of arrows:

    | *fork* ::= **if** *guarded_arrow*
    |        {{ **elsif** *guarded_arrow* }}
    |        [[ **else** *arrow* ]]
    |        **end**
    """

    def __init__(
        self,
        if_arrow: Arrow,
        elsif_arrows: Optional[List[Arrow]] = None,
        else_arrow: Optional[Arrow] = None,
    ) -> None:
        super().__init__()
        self._if_arrow = if_arrow
        self._elsif_arrows = elsif_arrows if elsif_arrows else []
        self._else_arrow = else_arrow

    @property
    def if_arrow(self) -> Arrow:
        """Start arrow."""
        return self._if_arrow

    @property
    def elsif_arrows(self) -> List[Arrow]:
        """Elsif arrows list."""
        return self._elsif_arrows

    @property
    def else_arrow(self) -> Union[Arrow, None]:
        """Else arrow."""
        return self._else_arrow


class ForkWithPriority(common.SwanItem):  # numpydoc ignore=PR01
    """Fork as a priority fork declaration:

    | *fork_priority* ::= *priority* **if** *guarded_arrow*
    |                  | *priority **else** *arrow*
    """

    def __init__(self, priority: Union[Literal, None], arrow: Arrow, is_if_arrow: bool) -> None:
        super().__init__()
        self._priority = priority
        self._arrow = arrow
        self._is_if_arrow = is_if_arrow

    @property
    def priority(self) -> Union[Literal, None]:
        """Fork priority."""
        return self._priority

    @property
    def arrow(self) -> Arrow:
        """For arrow."""
        return self._arrow

    @property
    def is_if_arrow(self) -> bool:
        """True when fork is *priority* **if** *guarded_arrow*,
        False if fork is *priority* **else** *arrow*.
        """
        return self._is_if_arrow

    @property
    def is_valid(self) -> bool:
        """Check if fork is either an **if** with a *guarded_arrow*, or
        an **else** with an *arrow*."""
        if self.is_if_arrow:
            return self.arrow.guard is not None
        return self.arrow.guard is None


class ForkPriorityList(Fork):  # numpydoc ignore=PR01
    """List of :py:class:`ForkWithPriority`.

    *fork* ::=  {{ *fork_priority* }} **end**
    """

    def __init__(self, prio_forks: List[ForkWithPriority]) -> None:
        super().__init__()
        self._prio_forks = prio_forks

    @property
    def prio_forks(self) -> List[ForkWithPriority]:
        """List of fork with priority."""
        return self._prio_forks


class Transition(common.SwanItem, common.PragmaBase):  # numpydoc ignore=PR01
    """State machine transition:

    | *transition* ::= **if** *guarded_arrow* ;
    |              | [[ *scope* ]] *target* ;

    """

    def __init__(self, arrow: Arrow, pragmas: Optional[List[common.Pragma]] = None) -> None:
        common.SwanItem.__init__(self)
        common.PragmaBase.__init__(self, pragmas)
        self._arrow = arrow

    @property
    def arrow(self) -> Arrow:
        """Transition arrow."""
        return self._arrow

    @property
    def is_guarded(self) -> bool:
        """True when arrow is guarded."""
        return self.arrow.guard is not None


class TransitionDecl(StateMachineItem):  # numpydoc ignore=PR01
    """Declaration of state machine transition:

    | *transition_decl* ::= *priority* [[ *state_ref* ]]
    |                      (( **unless** | **until** )) *transition*
    | *priority* ::= : [[ INTEGER ]] :

    """

    def __init__(
        self,
        priority: Union[Literal, None],
        transition: Transition,
        is_strong: bool,
        state_ref: StateRef,
    ) -> None:
        super().__init__()
        self._priority = priority
        self._transition = transition
        self._is_strong = is_strong
        self._state_ref = state_ref
        transition.owner = self

    @property
    def priority(self) -> Union[Literal, None]:
        """Transition priority."""
        return self._priority

    @property
    def transition(self) -> Transition:
        """Transition data."""
        return self._transition

    @property
    def is_strong(self) -> bool:
        """True when strong transition, else weak transition."""
        return self._is_strong

    @property
    def state_ref(self) -> StateRef:
        return self._state_ref


class State(StateMachineItem, common.PragmaBase):  # numpydoc ignore=PR01
    """State definition."""

    def __init__(
        self,
        id: Optional[common.Identifier] = None,
        lunum: Optional[common.Lunum] = None,
        strong_transitions: Optional[List[Transition]] = None,
        sections: Optional[List[scopes.ScopeSection]] = None,
        weak_transitions: Optional[List[Transition]] = None,
        is_initial: Optional[bool] = False,
        pragmas: Optional[List[common.Pragma]] = None,
    ) -> None:
        StateMachineItem.__init__(self)
        common.PragmaBase.__init__(self, pragmas)
        self._id = id
        self._lunum = lunum
        self._strong_transitions = strong_transitions if strong_transitions else []
        self._sections = sections if sections else []
        self._weak_transitions = weak_transitions if weak_transitions else []
        self._is_initial = is_initial
        common.SwanItem.set_owner(self, self._strong_transitions)
        common.SwanItem.set_owner(self, self._weak_transitions)
        common.SwanItem.set_owner(self, sections)

    @property
    def id(self) -> Union[common.Identifier, None]:
        return self._id

    @property
    def lunum(self) -> Union[common.Lunum, None]:
        return self._lunum

    @property
    def strong_transitions(self) -> List[Transition]:
        return self._strong_transitions

    @property
    def has_strong_transition(self) -> bool:
        """True when state has strong transitions."""
        return True if self.strong_transitions else False

    @property
    def sections(self) -> List[scopes.ScopeSection]:
        return self._sections

    @property
    def has_body(self) -> bool:
        """True when state has a body, namely scope sections."""
        return True if self.sections else False

    @property
    def weak_transitions(self) -> List[Transition]:
        return self._weak_transitions

    @property
    def has_weak_transition(self) -> bool:
        """True when state has weak transitions."""
        return True if self.weak_transitions else False

    @property
    def is_initial(self) -> Optional[bool]:
        return self._is_initial


class StateMachine(DefByCase):  # numpydoc ignore=PR01
    """State machine definition."""

    def __init__(
        self,
        lhs: Optional[EquationLHS] = None,
        items: Optional[List[StateMachineItem]] = None,
        name: Optional[common.Luid] = None,
        is_equation: bool = False,
    ) -> None:
        super().__init__(lhs, name, is_equation)
        self._items = items if items else []
        common.SwanItem.set_owner(self, self._items)

    @property
    def items(self) -> List[StateMachineItem]:
        """Transitions and states of the state machine."""
        return self._items


#
# Activates
# ============================================================

# Activate If
# ------------------------------------------------------------


class IfteBranch(common.SwanItem):  # numpydoc ignore=PR01
    """
    Base class for :py:class:`IfteDataDef` and :py:class:`IfteIfActivation` classes.

    | ifte_branch ::= data_def
    |             | if_activation
    """

    def __init__(self) -> None:
        super().__init__()


class IfActivationBranch(common.SwanItem):  # numpydoc ignore=PR01
    """Stores a branch of an *if_activation*.

    A branch is:

    - **if** *expr* **then** *ifte_branch*, or
    - **elsif** *expr* **then** *ifte_branch*, or
    - **else** *ifte_branch*

    """

    def __init__(self, condition: Union[common.Expression, None], branch: IfteBranch) -> None:
        super().__init__()
        self._condition = condition
        self._branch = branch
        common.SwanItem.set_owner(self, branch)

    @property
    def condition(self) -> Union[common.Expression, None]:
        """Branch condition, None for **else** branch."""
        return self._condition

    @property
    def branch(self) -> IfteBranch:
        """Branch activation branch."""
        return self._branch


class IfActivation(common.SwanItem):  # numpydoc ignore=PR01
    """
    List of *if_activation* branches as a list of :py:class:`IfActivationBranch`.

    | *if_activation* ::= **if** *expr* **then** *ifte_branch*
    |                     {{ **elsif** *expr* **then** *ifte_branch* }}
    |                     **else** *ifte_branch*
    """

    def __init__(self, branches: List[IfActivationBranch]) -> None:
        super().__init__()
        self._branches = branches
        common.SwanItem.set_owner(self, branches)

    @property
    def branches(self) -> List[IfActivationBranch]:
        """Return branches of *if_activation*.
        There must be at least two branches, the **if** and the **else** branches."""
        return self._branches

    @property
    def is_valid(self) -> bool:
        """Activation branches must be at least **if** and **else**, and *elsif* has a condition."""
        if len(self.branches) < 2:
            return False
        if self.branches[0].condition is None:
            return False
        if self.branches[-1].condition is None:
            return False
        # check all elsif as non None condition
        if len(self.branches) > 2:
            non_cond = list(filter(lambda x: x.condition is None, [self.branches[1:-1]]))
            if non_cond:
                return False
        return True


class IfteDataDef(IfteBranch):  # numpydoc ignore=PR01
    """
    *ifte_branch* of an **activate if** as a data definition. See :py:class:`ActivateIf`.

    *ifte_branch* ::= *data_def*
    """

    def __init__(self, data_def: Union[common.Equation, scopes.Scope]) -> None:
        super().__init__()
        self._data_def = data_def
        common.SwanItem.set_owner(self, data_def)

    @property
    def data_def(self) -> Union[common.Equation, scopes.Scope]:
        return self._data_def


class IfteIfActivation(IfteBranch):  # numpydoc ignore=PR01
    """
    *ifte_branch* of an **activate if** as an *if_activation*. See :py:class:`ActivateIf`.

    *ifte_branch* ::= *if_activation*
    """

    def __init__(self, if_activation: IfActivation) -> None:
        super().__init__()
        self._if_activation = if_activation
        common.SwanItem.set_owner(self, if_activation)

    @property
    def if_activation(self) -> IfActivation:
        """If activation."""
        return self._if_activation


class ActivateIf(DefByCase):  # numpydoc ignore=PR01
    """Activate if operator definition:

    | *select_activation* ::= **activate** [[ LUID ]] *if_activation*
    | *if_activation* ::= **if** *expr* **then** *ifte_branch*
    |                     {{ **elsif** *expr* **then** *ifte_branch* }}
    |                     **else** *ifte_branch*
    | *ifte_branch* ::= *data_def* | *if_activation*
    """

    def __init__(
        self,
        if_activation: IfActivation,
        lhs: Optional[EquationLHS] = None,
        name: Optional[common.Luid] = None,
        is_equation: bool = False,
    ) -> None:
        super().__init__(lhs, name, is_equation)
        self._if_activation = if_activation
        common.SwanItem.set_owner(self, if_activation)

    @property
    def if_activation(self) -> IfActivation:
        """Activation branch of **activate**."""
        return self._if_activation


# Activate When
# ------------------------------------------------------------
class ActivateWhenBranch(common.SwanItem):  # numpydoc ignore=PR01
    """Stores a branch of a *match_activation*.

    A branch is:
    **|** *pattern_with_capture* : *data_def*

    """

    def __init__(self, pattern: Pattern, data_def: Union[common.Equation, scopes.Scope]) -> None:
        super().__init__()
        self._pattern = pattern
        self._data_def = data_def

    @property
    def pattern(self) -> Pattern:
        """Branch pattern."""
        return self._pattern

    @property
    def data_def(self) -> Union[common.Equation, scopes.Scope]:
        """Branch data definition."""
        return self._data_def


class ActivateWhen(DefByCase):  # numpydoc ignore=PR01
    """Activate when operator definition.

    There must be at least one branch.
    This can be checked with the *is_valid()* method.

    | *select_activation* ::= *activate* [[ LUID ]] *match_activation*
    | *match_activation* ::= **when** *expr* **match**
    |                      {{ | *pattern_with_capture* : *data_def* }}+
    """

    def __init__(
        self,
        condition: common.Expression,
        branches: List[ActivateWhenBranch],
        lhs: Optional[EquationLHS] = None,
        name: Optional[str] = None,
        is_equation: bool = False,
    ) -> None:
        super().__init__(lhs, name, is_equation)
        self._condition = condition
        self._branches = branches
        common.SwanItem.set_owner(self, condition)
        common.SwanItem.set_owner(self, branches)

    @property
    def is_valid(self) -> bool:
        """True when there is at least one branch."""
        return len(self.branches) > 0

    @property
    def condition(self) -> common.Expression:
        """Activate when condition."""
        return self._condition

    @property
    def branches(self) -> List[ActivateWhenBranch]:
        """Activate when branches."""
        return self._branches



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from abc import ABC, abstractmethod
from typing import Union

from ansys.scadeone.core.common.storage import SwanStorage
from ansys.scadeone.core.common.exception import ScadeOneException
import ansys.scadeone.core.swan as S


class Parser(ABC):
    """The parser base class as a proxy to the F# methods implemented
    by the F# parser.
    """

    # Current parser in used. Set by derived class
    _CurrentParser = None

    @classmethod
    def get_current_parser(cls) -> "Parser":
        """Returns the current parser in use. Parser must be set."""
        if cls._CurrentParser:
            return cls._CurrentParser
        raise ScadeOneException("Current parser not set.")

    @classmethod
    def set_current_parser(cls, parser: "Parser"):
        cls._CurrentParser = parser

    # Source of parsing
    _SwanSource = None

    @classmethod
    def get_source(cls) -> SwanStorage:
        return cls._SwanSource

    @classmethod
    def set_source(cls, swan: SwanStorage) -> SwanStorage:
        cls._SwanSource = swan

    @abstractmethod
    def module_body(self, source: SwanStorage) -> S.ModuleBody:
        """Parse a Swan module from a SwanStorage object

            The *content()* method is called to get the code.

            The *name* property is used to set the module identifier.

        Parameters
        ----------
        source : SwanStorage
            Swan module (.swan)

        Returns
        -------
        ModuleBody:
            Instance of a module body
        """
        pass

    @abstractmethod
    def module_interface(self, source: SwanStorage) -> S.ModuleInterface:
        """Parse a Swan interface from a SwanStorage object.

            The *content()* method is called to get the code.

            The *name* property is used to set the module identifier.

        Parameters
        ----------
        source : SwanStorage
            Swan interface (.swani)

        Returns
        -------
        ModuleInterface
            Instance of a module interface.
        """
        pass

    @abstractmethod
    def declaration(self, source: SwanStorage) -> S.Declaration:
        """Parse a Swan declaration:
          type, const, sensor, group, use, operator (signature or with body).

        Parameters
        ----------
        source : SwanStorage
            Single Swan declaration

        Returns
        -------
        Declaration
            Instance Declaration object
        """
        pass

    @abstractmethod
    def equation(self, source: SwanStorage) -> S.equations:
        """Parse a Swan equation.

        Parameters
        ----------
        source : SwanStorage
            Swan equation text

        Returns
        -------
        Equation
            Instance of Equation object
        """
        pass

    @abstractmethod
    def expression(self, source: SwanStorage) -> S.Expression:
        """Parse a Swan expression

        Parameters
        ----------
        source : SwanStorage
            Swan expression text

        Returns
        -------
        Expression
            Instance of an expression object
        """

    @abstractmethod
    def scope_section(self, source: SwanStorage) -> S.ScopeSection:
        """Parse a Swan scope section

        Parameters
        ----------
        source : str
            Swan scope section text

        Returns
        -------
        ScopeSection
            Instance of a scope section object
        """
        pass

    @abstractmethod
    def op_expr(self, source: SwanStorage) -> S.OperatorExpression:
        """Parse a Swan operator expression

        Parameters
        ----------
        source : SwanStorage
            Swan code for operator expression

        Returns
        -------
        OperatorExpression
            Instance of the operator expression object
        """
        pass

    @abstractmethod
    def operator_block(self, source: SwanStorage) -> Union[S.OperatorBase, S.OperatorExpression]:
        """Parse a Swan operator block

        *operator_block* ::= *operator* | *op_expr*

        Parameters
        ----------
        source : SwanStorage
            Swan code for operator block

        Returns
        -------
        Union[S.Operator, S.OperatorExpression]
            Instance of the *operator* or *op_expr*
        """
        pass

    @abstractmethod
    def operator_decl(self, source: SwanStorage) -> Union[S.Operator, S.Signature]:
        """Parse a Swan operator

        Parameters
        ----------
        source : SwanStorage
            Swan operator text or operator interface

        Returns
        -------
        S.Operator|S.Signature
            Instance of the operator, or its signature
        """
        pass



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylint: disable=too-many-lines, pointless-statement, invalid-name
from abc import ABC
from typing import Any, Optional

from ansys.scadeone.core import swan

__all__ = ["SwanVisitor", "Owner", "OwnerProperty"]

Owner = Optional[swan.SwanItem]
OwnerProperty = Optional[str]


class SwanVisitor(ABC):
    """Visitor for Swan classes. This class must be derived."""

    @staticmethod
    def _is_builtin(obj: Any) -> bool:
        """Return True if name is a simple builtin-type
        that can be found in swan modules"""
        return type(obj).__name__ in ("str", "int", "bool", "float")

    def visit(self, swan_obj: swan.SwanItem) -> None:
        """Entry point. After creation of an instance of
        a SwanVisitor, this method must be called on a Swan object instance."""
        self._visit(swan_obj, None, None)

    def _visit(
        self,
        swan_obj: swan.SwanItem,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Dispatch function. Visit a Swan object and its properties recursively
        by calling `self.visit_<swan_obj class name>(swan_obj, owner, owner_property)`.

        Parameters
        ----------
        swan_obj : swan.SwanItem
            Visited Swan object.
        owner : Owner
            Owner of the swan_obj, 'None' for the root visited object
        owner_property : OwnerProperty
            Owner property name to know the visit context, 'None' for the root visited object
        """
        try:
            class_name = swan_obj.__class__.__name__
            fn = getattr(self, f"visit_{class_name}")
        except AttributeError:
            raise AttributeError(f"{self.__class__.__name__}: no visitor for {type(swan_obj)}")
        fn(swan_obj, owner, owner_property)

    # Following methods must be overridden

    def visit_builtin(
        self,
        object: Any,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Called when visiting a builtin value: `str`, `bool`, `int` or `float`.
        Override this method."""
        pass

    def visit_SwanItem(
        self,
        swan_obj: swan.SwanItem,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Visit of the base class for every Swan constructs.
        Override this method."""
        pass

    # Class visitors
    def visit_ActivateClock(
        self,
        swan_obj: swan.ActivateClock,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ActivateClock visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_OperatorExpression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.operator, swan_obj, "operator")
        self._visit(swan_obj.clock, swan_obj, "clock")

    def visit_ActivateEvery(
        self,
        swan_obj: swan.ActivateEvery,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ActivateEvery visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_OperatorExpression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.operator, swan_obj, "operator")
        self._visit(swan_obj.condition, swan_obj, "condition")
        self.visit_builtin(swan_obj.is_last, swan_obj, "is_last")
        self._visit(swan_obj.expr, swan_obj, "expr")

    def visit_ActivateIf(
        self,
        swan_obj: swan.ActivateIf,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ActivateIf visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_DefByCase(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.if_activation, swan_obj, "if_activation")

    def visit_ActivateIfBlock(
        self,
        swan_obj: swan.ActivateIfBlock,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ActivateIfBlock visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_DefByCaseBlockBase(swan_obj, owner, owner_property)

    def visit_ActivateWhen(
        self,
        swan_obj: swan.ActivateWhen,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ActivateWhen visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_DefByCase(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.condition, swan_obj, "condition")
        for item in swan_obj.branches:
            self._visit(item, swan_obj, "branches")

    def visit_ActivateWhenBlock(
        self,
        swan_obj: swan.ActivateWhenBlock,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ActivateWhenBlock visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_DefByCaseBlockBase(swan_obj, owner, owner_property)

    def visit_ActivateWhenBranch(
        self,
        swan_obj: swan.ActivateWhenBranch,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ActivateWhenBranch visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.pattern, swan_obj, "pattern")
        if isinstance(swan_obj.data_def, swan.Equation):
            self._visit(swan_obj.data_def, swan_obj, "data_def")
        elif isinstance(swan_obj.data_def, swan.Scope):
            self._visit(swan_obj.data_def, swan_obj, "data_def")

    def visit_AnonymousOperatorWithDataDefinition(
        self,
        swan_obj: swan.AnonymousOperatorWithDataDefinition,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """AnonymousOperatorWithDataDefinition visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_OperatorExpression(swan_obj, owner, owner_property)
        # Visit properties
        self.visit_builtin(swan_obj.is_node, swan_obj, "is_node")
        for item in swan_obj.inputs:
            self._visit(item, swan_obj, "inputs")
        for item in swan_obj.outputs:
            self._visit(item, swan_obj, "outputs")
        if isinstance(swan_obj.data_def, swan.Equation):
            self._visit(swan_obj.data_def, swan_obj, "data_def")
        elif isinstance(swan_obj.data_def, swan.Scope):
            self._visit(swan_obj.data_def, swan_obj, "data_def")

    def visit_AnonymousOperatorWithExpression(
        self,
        swan_obj: swan.AnonymousOperatorWithExpression,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """AnonymousOperatorWithExpression visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_OperatorExpression(swan_obj, owner, owner_property)
        # Visit properties
        self.visit_builtin(swan_obj.is_node, swan_obj, "is_node")
        for item in swan_obj.params:
            self._visit(item, swan_obj, "params")
        for item in swan_obj.sections:
            self._visit(item, swan_obj, "sections")
        self._visit(swan_obj.expr, swan_obj, "expr")

    def visit_ArrayConstructor(
        self,
        swan_obj: swan.ArrayConstructor,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ArrayConstructor visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.group, swan_obj, "group")

    def visit_ArrayProjection(
        self,
        swan_obj: swan.ArrayProjection,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ArrayProjection visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.expr, swan_obj, "expr")
        self._visit(swan_obj.index, swan_obj, "index")

    def visit_ArrayRepetition(
        self,
        swan_obj: swan.ArrayRepetition,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ArrayRepetition visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.expr, swan_obj, "expr")
        self._visit(swan_obj.size, swan_obj, "size")

    def visit_ArrayTypeExpression(
        self,
        swan_obj: swan.ArrayTypeExpression,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ArrayTypeExpression visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_TypeExpression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.type, swan_obj, "type")
        self._visit(swan_obj.size, swan_obj, "size")

    def visit_Arrow(
        self,
        swan_obj: swan.Arrow,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Arrow visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.guard is not None:
            self._visit(swan_obj.guard, swan_obj, "guard")
        if swan_obj.action is not None:
            self._visit(swan_obj.action, swan_obj, "action")
        if swan_obj.target is not None:
            self._visit(swan_obj.target, swan_obj, "target")
        if swan_obj.fork is not None:
            self._visit(swan_obj.fork, swan_obj, "fork")

    def visit_AssertSection(
        self,
        swan_obj: swan.AssertSection,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """AssertSection visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_ScopeSection(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.assertions:
            self._visit(item, swan_obj, "assertions")

    def visit_AssumeSection(
        self,
        swan_obj: swan.AssumeSection,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """AssumeSection visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_ScopeSection(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.hypotheses:
            self._visit(item, swan_obj, "hypotheses")

    def visit_Bar(
        self,
        swan_obj: swan.Bar,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Bar visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_DiagramObject(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.operation is not None:
            self._visit(swan_obj.operation, swan_obj, "operation")

    def visit_BinaryExpr(
        self,
        swan_obj: swan.BinaryExpr,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """BinaryExpr visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.operator, swan_obj, "operator")
        self._visit(swan_obj.left, swan_obj, "left")
        self._visit(swan_obj.right, swan_obj, "right")

    def visit_BinaryOp(
        self,
        swan_obj: swan.BinaryOp,
        owner: Owner,
        owner_property: OwnerProperty,
    ):
        """BinaryOp visitor function. Should be overridden."""
        # Enum values:
        # Plus
        # Minus
        # Mult
        # Slash
        # Mod
        # Equal
        # Diff
        # Lt
        # Gt
        # Leq
        # Geq
        # And
        # Or
        # Xor
        # Land
        # Lor
        # Lxor
        # Lsl
        # Lsr
        # Arrow
        # Pre
        # Concat
        pass

    def visit_Block(
        self,
        swan_obj: swan.Block,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Block visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_DiagramObject(swan_obj, owner, owner_property)
        # Visit properties
        if isinstance(swan_obj.instance, swan.OperatorBase):
            self._visit(swan_obj.instance, swan_obj, "instance")
        elif isinstance(swan_obj.instance, swan.OperatorExpression):
            self._visit(swan_obj.instance, swan_obj, "instance")
        elif isinstance(swan_obj.instance, swan.ProtectedItem):
            self._visit(swan_obj.instance, swan_obj, "instance")

    def visit_BoolPattern(
        self,
        swan_obj: swan.BoolPattern,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """BoolPattern visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Pattern(swan_obj, owner, owner_property)
        # Visit properties
        self.visit_builtin(swan_obj.value, swan_obj, "value")

    def visit_BoolType(
        self,
        swan_obj: swan.BoolType,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """BoolType visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_PredefinedType(swan_obj, owner, owner_property)

    def visit_CaseBranch(
        self,
        swan_obj: swan.CaseBranch,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """CaseBranch visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.pattern, swan_obj, "pattern")
        self._visit(swan_obj.expr, swan_obj, "expr")

    def visit_CaseExpr(
        self,
        swan_obj: swan.CaseExpr,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """CaseExpr visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.expr, swan_obj, "expr")
        for item in swan_obj.branches:
            self._visit(item, swan_obj, "branches")

    def visit_CharPattern(
        self,
        swan_obj: swan.CharPattern,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """CharPattern visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Pattern(swan_obj, owner, owner_property)
        # Visit properties
        self.visit_builtin(swan_obj.value, swan_obj, "value")

    def visit_CharType(
        self,
        swan_obj: swan.CharType,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """CharType visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_PredefinedType(swan_obj, owner, owner_property)

    def visit_ClockExpr(
        self,
        swan_obj: swan.ClockExpr,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ClockExpr visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.id, swan_obj, "id")
        if swan_obj.is_not is not None:
            self.visit_builtin(swan_obj.is_not, swan_obj, "is_not")
        if swan_obj.pattern is not None:
            self._visit(swan_obj.pattern, swan_obj, "pattern")

    def visit_Connection(
        self,
        swan_obj: swan.Connection,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Connection visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.port is not None:
            self._visit(swan_obj.port, swan_obj, "port")
        if swan_obj.adaptation is not None:
            self._visit(swan_obj.adaptation, swan_obj, "adaptation")

    def visit_ConstDecl(
        self,
        swan_obj: swan.ConstDecl,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ConstDecl visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Declaration(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.type is not None:
            self._visit(swan_obj.type, swan_obj, "type")
        if swan_obj.value is not None:
            self._visit(swan_obj.value, swan_obj, "value")

    def visit_ConstDeclarations(
        self,
        swan_obj: swan.ConstDeclarations,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ConstDeclarations visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_GlobalDeclaration(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.constants:
            self._visit(item, swan_obj, "constants")

    def visit_Declaration(
        self,
        swan_obj: swan.Declaration,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Declaration visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.id, swan_obj, "id")

    def visit_DefBlock(
        self,
        swan_obj: swan.DefBlock,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """DefBlock visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_DiagramObject(swan_obj, owner, owner_property)
        # Visit properties
        if isinstance(swan_obj.lhs, swan.EquationLHS):
            self._visit(swan_obj.lhs, swan_obj, "lhs")
        elif isinstance(swan_obj.lhs, swan.ProtectedItem):
            self._visit(swan_obj.lhs, swan_obj, "lhs")

    def visit_DefByCase(
        self,
        swan_obj: swan.DefByCase,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """DefByCase visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Equation(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.lhs is not None:
            self._visit(swan_obj.lhs, swan_obj, "lhs")
        if swan_obj.name is not None:
            self._visit(swan_obj.name, swan_obj, "name")
        self.visit_builtin(swan_obj.is_equation, swan_obj, "is_equation")

    def visit_DefByCaseBlockBase(
        self,
        swan_obj: swan.DefByCaseBlockBase,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """DefByCaseBlockBase visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_DiagramObject(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.def_by_case, swan_obj, "def_by_case")

    def visit_DefaultPattern(
        self,
        swan_obj: swan.DefaultPattern,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """DefaultPattern visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Pattern(swan_obj, owner, owner_property)

    def visit_Diagram(
        self,
        swan_obj: swan.Diagram,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Diagram visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_ScopeSection(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.objects:
            self._visit(item, swan_obj, "objects")

    def visit_DiagramObject(
        self,
        swan_obj: swan.DiagramObject,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """DiagramObject visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        self.visit_PragmaBase(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.lunum is not None:
            self._visit(swan_obj.lunum, swan_obj, "lunum")
        if swan_obj.luid is not None:
            self._visit(swan_obj.luid, swan_obj, "luid")
        if swan_obj.locals is not None:
            for item in swan_obj.locals:
                self._visit(item, swan_obj, "locals")

    def visit_EmissionBody(
        self,
        swan_obj: swan.EmissionBody,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """EmissionBody visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.flows:
            self._visit(item, swan_obj, "flows")
        if swan_obj.condition is not None:
            self._visit(swan_obj.condition, swan_obj, "condition")
        if swan_obj.luid is not None:
            self._visit(swan_obj.luid, swan_obj, "luid")

    def visit_EmitSection(
        self,
        swan_obj: swan.EmitSection,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """EmitSection visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_ScopeSection(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.emissions:
            self._visit(item, swan_obj, "emissions")

    def visit_EnumTypeDefinition(
        self,
        swan_obj: swan.EnumTypeDefinition,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """EnumTypeDefinition visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_TypeDefinition(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.tags:
            self._visit(item, swan_obj, "tags")

    def visit_Equation(
        self,
        swan_obj: swan.Equation,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Equation visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)

    def visit_EquationLHS(
        self,
        swan_obj: swan.EquationLHS,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """EquationLHS visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.lhs_items:
            self._visit(item, swan_obj, "lhs_items")
        if swan_obj.is_partial_lhs is not None:
            self.visit_builtin(swan_obj.is_partial_lhs, swan_obj, "is_partial_lhs")

    def visit_ExprBlock(
        self,
        swan_obj: swan.ExprBlock,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ExprBlock visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_DiagramObject(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.expr, swan_obj, "expr")

    def visit_ExprEquation(
        self,
        swan_obj: swan.ExprEquation,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ExprEquation visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Equation(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.lhs, swan_obj, "lhs")
        self._visit(swan_obj.expr, swan_obj, "expr")
        if swan_obj.luid is not None:
            self._visit(swan_obj.luid, swan_obj, "luid")

    def visit_ExprTypeDefinition(
        self,
        swan_obj: swan.ExprTypeDefinition,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ExprTypeDefinition visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_TypeDefinition(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.type, swan_obj, "type")

    def visit_Expression(
        self,
        swan_obj: swan.Expression,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Expression visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)

    def visit_Float32Type(
        self,
        swan_obj: swan.Float32Type,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Float32Type visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_PredefinedType(swan_obj, owner, owner_property)

    def visit_Float64Type(
        self,
        swan_obj: swan.Float64Type,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Float64Type visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_PredefinedType(swan_obj, owner, owner_property)

    def visit_Fork(
        self,
        swan_obj: swan.Fork,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Fork visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)

    def visit_ForkPriorityList(
        self,
        swan_obj: swan.ForkPriorityList,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ForkPriorityList visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Fork(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.prio_forks:
            self._visit(item, swan_obj, "prio_forks")

    def visit_ForkTree(
        self,
        swan_obj: swan.ForkTree,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ForkTree visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Fork(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.if_arrow, swan_obj, "if_arrow")
        if swan_obj.elsif_arrows is not None:
            for item in swan_obj.elsif_arrows:
                self._visit(item, swan_obj, "elsif_arrows")
        if swan_obj.else_arrow is not None:
            self._visit(swan_obj.else_arrow, swan_obj, "else_arrow")

    def visit_ForkWithPriority(
        self,
        swan_obj: swan.ForkWithPriority,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ForkWithPriority visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.priority is not None:
            self._visit(swan_obj.priority, swan_obj, "priority")
        self._visit(swan_obj.arrow, swan_obj, "arrow")
        self.visit_builtin(swan_obj.is_if_arrow, swan_obj, "is_if_arrow")

    def visit_FormalProperty(
        self,
        swan_obj: swan.FormalProperty,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """FormalProperty visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.luid, swan_obj, "luid")
        self._visit(swan_obj.expr, swan_obj, "expr")

    def visit_Forward(
        self,
        swan_obj: swan.Forward,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Forward visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.state, swan_obj, "state")
        for item in swan_obj.dimensions:
            self._visit(item, swan_obj, "dimensions")
        self._visit(swan_obj.body, swan_obj, "body")
        for item in swan_obj.returns:
            self._visit(item, swan_obj, "returns")
        if swan_obj.luid is not None:
            self._visit(swan_obj.luid, swan_obj, "luid")

    def visit_ForwardArrayClause(
        self,
        swan_obj: swan.ForwardArrayClause,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ForwardArrayClause visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        if isinstance(swan_obj.return_clause, swan.ForwardItemClause):
            self._visit(swan_obj.return_clause, swan_obj, "return_clause")
        elif isinstance(swan_obj.return_clause, swan.ForwardArrayClause):
            self._visit(swan_obj.return_clause, swan_obj, "return_clause")

    def visit_ForwardBody(
        self,
        swan_obj: swan.ForwardBody,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ForwardBody visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.body:
            self._visit(item, swan_obj, "body")
        if swan_obj.unless_expr is not None:
            self._visit(swan_obj.unless_expr, swan_obj, "unless_expr")
        if swan_obj.until_expr is not None:
            self._visit(swan_obj.until_expr, swan_obj, "until_expr")

    def visit_ForwardDim(
        self,
        swan_obj: swan.ForwardDim,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ForwardDim visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.expr is not None:
            self._visit(swan_obj.expr, swan_obj, "expr")
        if swan_obj.dim_id is not None:
            self._visit(swan_obj.dim_id, swan_obj, "dim_id")
        if swan_obj.elems is not None:
            for item in swan_obj.elems:
                self._visit(item, swan_obj, "elems")
        if swan_obj.protected is not None:
            self.visit_builtin(swan_obj.protected, swan_obj, "protected")

    def visit_ForwardElement(
        self,
        swan_obj: swan.ForwardElement,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ForwardElement visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.lhs, swan_obj, "lhs")
        self._visit(swan_obj.expr, swan_obj, "expr")

    def visit_ForwardItemClause(
        self,
        swan_obj: swan.ForwardItemClause,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ForwardItemClause visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.id, swan_obj, "id")
        if swan_obj.last_default is not None:
            self._visit(swan_obj.last_default, swan_obj, "last_default")

    def visit_ForwardLHS(
        self,
        swan_obj: swan.ForwardLHS,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ForwardLHS visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        if isinstance(swan_obj.lhs, swan.Identifier):
            self._visit(swan_obj.lhs, swan_obj, "lhs")
        elif isinstance(swan_obj.lhs, swan.ForwardLHS):
            self._visit(swan_obj.lhs, swan_obj, "lhs")

    def visit_ForwardLastDefault(
        self,
        swan_obj: swan.ForwardLastDefault,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ForwardLastDefault visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.last is not None:
            self._visit(swan_obj.last, swan_obj, "last")
        if swan_obj.default is not None:
            self._visit(swan_obj.default, swan_obj, "default")
        if swan_obj.shared is not None:
            self._visit(swan_obj.shared, swan_obj, "shared")

    def visit_ForwardReturnArrayClause(
        self,
        swan_obj: swan.ForwardReturnArrayClause,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ForwardReturnArrayClause visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_ForwardReturnItem(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.array_clause, swan_obj, "array_clause")
        if swan_obj.return_id is not None:
            self._visit(swan_obj.return_id, swan_obj, "return_id")

    def visit_ForwardReturnItem(
        self,
        swan_obj: swan.ForwardReturnItem,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ForwardReturnItem visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)

    def visit_ForwardReturnItemClause(
        self,
        swan_obj: swan.ForwardReturnItemClause,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ForwardReturnItemClause visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_ForwardReturnItem(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.item_clause, swan_obj, "item_clause")

    def visit_ForwardState(
        self,
        swan_obj: swan.ForwardState,
        owner: Owner,
        owner_property: OwnerProperty,
    ):
        """ForwardState visitor function. Should be overridden."""
        # Enum values:
        # Nothing
        # Restart
        # Resume
        pass

    def visit_FunctionalUpdate(
        self,
        swan_obj: swan.FunctionalUpdate,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """FunctionalUpdate visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.expr, swan_obj, "expr")
        for item in swan_obj.modifiers:
            self._visit(item, swan_obj, "modifiers")

    def visit_GlobalDeclaration(
        self,
        swan_obj: swan.GlobalDeclaration,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """GlobalDeclaration visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_ModuleItem(swan_obj, owner, owner_property)

    def visit_Group(
        self,
        swan_obj: swan.Group,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Group visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.items:
            self._visit(item, swan_obj, "items")

    def visit_GroupAdaptation(
        self,
        swan_obj: swan.GroupAdaptation,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """GroupAdaptation visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.renamings:
            self._visit(item, swan_obj, "renamings")

    def visit_GroupConstructor(
        self,
        swan_obj: swan.GroupConstructor,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """GroupConstructor visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.group, swan_obj, "group")

    def visit_GroupDecl(
        self,
        swan_obj: swan.GroupDecl,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """GroupDecl visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Declaration(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.type, swan_obj, "type")

    def visit_GroupDeclarations(
        self,
        swan_obj: swan.GroupDeclarations,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """GroupDeclarations visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_GlobalDeclaration(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.groups:
            self._visit(item, swan_obj, "groups")

    def visit_GroupItem(
        self,
        swan_obj: swan.GroupItem,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """GroupItem visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.expr, swan_obj, "expr")
        if swan_obj.label is not None:
            self._visit(swan_obj.label, swan_obj, "label")

    def visit_GroupOperation(
        self,
        swan_obj: swan.GroupOperation,
        owner: Owner,
        owner_property: OwnerProperty,
    ):
        """GroupOperation visitor function. Should be overridden."""
        # Enum values:
        # NoOp
        # ByName
        # ByPos
        # Normalize
        pass

    def visit_GroupProjection(
        self,
        swan_obj: swan.GroupProjection,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """GroupProjection visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.expr, swan_obj, "expr")
        self._visit(swan_obj.adaptation, swan_obj, "adaptation")

    def visit_GroupRenaming(
        self,
        swan_obj: swan.GroupRenaming,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """GroupRenaming visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_GroupRenamingBase(swan_obj, owner, owner_property)
        # Visit properties
        if isinstance(swan_obj.source, swan.Identifier):
            self._visit(swan_obj.source, swan_obj, "source")
        elif isinstance(swan_obj.source, swan.Literal):
            self._visit(swan_obj.source, swan_obj, "source")

        if swan_obj.renaming is not None:
            self._visit(swan_obj.renaming, swan_obj, "renaming")
        if swan_obj.is_shortcut is not None:
            self.visit_builtin(swan_obj.is_shortcut, swan_obj, "is_shortcut")

    def visit_GroupRenamingBase(
        self,
        swan_obj: swan.GroupRenamingBase,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """GroupRenamingBase visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)

    def visit_GroupTypeExpression(
        self,
        swan_obj: swan.GroupTypeExpression,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """GroupTypeExpression visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)

    def visit_GroupTypeExpressionList(
        self,
        swan_obj: swan.GroupTypeExpressionList,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """GroupTypeExpressionList visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_GroupTypeExpression(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.positional:
            self._visit(item, swan_obj, "positional")
        for item in swan_obj.named:
            self._visit(item, swan_obj, "named")

    def visit_GuaranteeSection(
        self,
        swan_obj: swan.GuaranteeSection,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """GuaranteeSection visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_ScopeSection(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.guarantees:
            self._visit(item, swan_obj, "guarantees")

    def visit_Identifier(
        self,
        swan_obj: swan.Identifier,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Identifier visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        self.visit_PragmaBase(swan_obj, owner, owner_property)
        # Visit properties
        self.visit_builtin(swan_obj.value, swan_obj, "value")
        if swan_obj.comment is not None:
            self.visit_builtin(swan_obj.comment, swan_obj, "comment")
        if swan_obj.is_name is not None:
            self.visit_builtin(swan_obj.is_name, swan_obj, "is_name")

    def visit_IfActivation(
        self,
        swan_obj: swan.IfActivation,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """IfActivation visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.branches:
            self._visit(item, swan_obj, "branches")

    def visit_IfActivationBranch(
        self,
        swan_obj: swan.IfActivationBranch,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """IfActivationBranch visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.condition is not None:
            self._visit(swan_obj.condition, swan_obj, "condition")
        self._visit(swan_obj.branch, swan_obj, "branch")

    def visit_IfteBranch(
        self,
        swan_obj: swan.IfteBranch,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """IfteBranch visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)

    def visit_IfteDataDef(
        self,
        swan_obj: swan.IfteDataDef,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """IfteDataDef visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_IfteBranch(swan_obj, owner, owner_property)
        # Visit properties
        if isinstance(swan_obj.data_def, swan.Equation):
            self._visit(swan_obj.data_def, swan_obj, "data_def")
        elif isinstance(swan_obj.data_def, swan.Scope):
            self._visit(swan_obj.data_def, swan_obj, "data_def")

    def visit_IfteExpr(
        self,
        swan_obj: swan.IfteExpr,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """IfteExpr visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.cond_expr, swan_obj, "cond_expr")
        self._visit(swan_obj.then_expr, swan_obj, "then_expr")
        self._visit(swan_obj.else_expr, swan_obj, "else_expr")

    def visit_IfteIfActivation(
        self,
        swan_obj: swan.IfteIfActivation,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """IfteIfActivation visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_IfteBranch(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.if_activation, swan_obj, "if_activation")

    def visit_Int16Type(
        self,
        swan_obj: swan.Int16Type,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Int16Type visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_PredefinedType(swan_obj, owner, owner_property)

    def visit_Int32Type(
        self,
        swan_obj: swan.Int32Type,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Int32Type visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_PredefinedType(swan_obj, owner, owner_property)

    def visit_Int64Type(
        self,
        swan_obj: swan.Int64Type,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Int64Type visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_PredefinedType(swan_obj, owner, owner_property)

    def visit_Int8Type(
        self,
        swan_obj: swan.Int8Type,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Int8Type visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_PredefinedType(swan_obj, owner, owner_property)

    def visit_IntPattern(
        self,
        swan_obj: swan.IntPattern,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """IntPattern visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Pattern(swan_obj, owner, owner_property)
        # Visit properties
        self.visit_builtin(swan_obj.value, swan_obj, "value")
        if swan_obj.is_minus is not None:
            self.visit_builtin(swan_obj.is_minus, swan_obj, "is_minus")

    def visit_Iterator(
        self,
        swan_obj: swan.Iterator,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Iterator visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_OperatorExpression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.kind, swan_obj, "kind")
        self._visit(swan_obj.operator, swan_obj, "operator")

    def visit_IteratorKind(
        self,
        swan_obj: swan.IteratorKind,
        owner: Owner,
        owner_property: OwnerProperty,
    ):
        """IteratorKind visitor function. Should be overridden."""
        # Enum values:
        # Map
        # Fold
        # Mapfold
        # Mapi
        # Foldi
        # Mapfoldi
        pass

    def visit_LHSItem(
        self,
        swan_obj: swan.LHSItem,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """LHSItem visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        if isinstance(swan_obj.id, swan.Identifier):
            self._visit(swan_obj.id, swan_obj, "id")
        elif SwanVisitor._is_builtin(swan_obj.id):
            self.visit_builtin(swan_obj.id, swan_obj, "id")

    def visit_LabelOrIndex(
        self,
        swan_obj: swan.LabelOrIndex,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """LabelOrIndex visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        if isinstance(swan_obj.value, swan.Identifier):
            self._visit(swan_obj.value, swan_obj, "value")
        elif isinstance(swan_obj.value, swan.Expression):
            self._visit(swan_obj.value, swan_obj, "value")

    def visit_LastExpr(
        self,
        swan_obj: swan.LastExpr,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """LastExpr visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.id, swan_obj, "id")

    def visit_LetSection(
        self,
        swan_obj: swan.LetSection,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """LetSection visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_ScopeSection(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.equations:
            self._visit(item, swan_obj, "equations")

    def visit_Literal(
        self,
        swan_obj: swan.Literal,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Literal visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self.visit_builtin(swan_obj.value, swan_obj, "value")

    def visit_LiteralKind(
        self,
        swan_obj: swan.LiteralKind,
        owner: Owner,
        owner_property: OwnerProperty,
    ):
        """LiteralKind visitor function. Should be overridden."""
        # Enum values:
        # Bool
        # Char
        # Numeric
        # Error
        pass

    def visit_Luid(
        self,
        swan_obj: swan.Luid,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Luid visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        self.visit_builtin(swan_obj.value, swan_obj, "value")

    def visit_Lunum(
        self,
        swan_obj: swan.Lunum,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Lunum visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        self.visit_builtin(swan_obj.value, swan_obj, "value")

    def visit_Merge(
        self,
        swan_obj: swan.Merge,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Merge visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.params:
            self._visit(item, swan_obj, "params")

    def visit_Modifier(
        self,
        swan_obj: swan.Modifier,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Modifier visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        if isinstance(swan_obj.modifier, list):
            for item in swan_obj.modifier:
                self._visit(item, swan_obj, "modifier")
        elif SwanVisitor._is_builtin(swan_obj.modifier):
            self.visit_builtin(swan_obj.modifier, swan_obj, "modifier")

        self._visit(swan_obj.expr, swan_obj, "expr")

    def visit_Module(
        self,
        swan_obj: swan.Module,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Module visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_ModuleBase(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.name, swan_obj, "name")
        for item in swan_obj.use_directives:
            self._visit(item, swan_obj, "use_directives")
        for item in swan_obj.declarations:
            self._visit(item, swan_obj, "declarations")

    def visit_ModuleBase(
        self,
        swan_obj: swan.ModuleBase,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ModuleBase visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)

    def visit_ModuleBody(
        self,
        swan_obj: swan.ModuleBody,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ModuleBody visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Module(swan_obj, owner, owner_property)

    def visit_ModuleInterface(
        self,
        swan_obj: swan.ModuleInterface,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ModuleInterface visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Module(swan_obj, owner, owner_property)

    def visit_ModuleItem(
        self,
        swan_obj: swan.ModuleItem,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ModuleItem visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)

    def visit_NAryOperator(
        self,
        swan_obj: swan.NAryOperator,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """NAryOperator visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_OperatorExpression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.operator, swan_obj, "operator")

    def visit_NamedGroupTypeExpression(
        self,
        swan_obj: swan.NamedGroupTypeExpression,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """NamedGroupTypeExpression visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_GroupTypeExpression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.label, swan_obj, "label")
        self._visit(swan_obj.type, swan_obj, "type")

    def visit_NaryOp(
        self,
        swan_obj: swan.NaryOp,
        owner: Owner,
        owner_property: OwnerProperty,
    ):
        """NaryOp visitor function. Should be overridden."""
        # Enum values:
        # Plus
        # Mult
        # Land
        # Lor
        # And
        # Or
        # Xor
        # Concat
        pass

    def visit_NumericCast(
        self,
        swan_obj: swan.NumericCast,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """NumericCast visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.expr, swan_obj, "expr")
        self._visit(swan_obj.type, swan_obj, "type")

    def visit_NumericKind(
        self,
        swan_obj: swan.NumericKind,
        owner: Owner,
        owner_property: OwnerProperty,
    ):
        """NumericKind visitor function. Should be overridden."""
        # Enum values:
        # Numeric
        # Integer
        # Signed
        # Unsigned
        # Float
        pass

    def visit_Operator(
        self,
        swan_obj: swan.Operator,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Operator visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_OperatorSignatureBase(swan_obj, owner, owner_property)
        # Visit properties
        if isinstance(swan_obj.body, swan.Scope):
            self._visit(swan_obj.body, swan_obj, "body")
        elif isinstance(swan_obj.body, swan.Equation):
            self._visit(swan_obj.body, swan_obj, "body")

    def visit_OperatorBase(
        self,
        swan_obj: swan.OperatorBase,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """OperatorBase visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.sizes:
            self._visit(item, swan_obj, "sizes")

    def visit_OperatorExpression(
        self,
        swan_obj: swan.OperatorExpression,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """OperatorExpression visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)

    def visit_OperatorInstance(
        self,
        swan_obj: swan.OperatorInstance,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """OperatorInstance visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.operator, swan_obj, "operator")
        self._visit(swan_obj.params, swan_obj, "params")
        if swan_obj.luid is not None:
            self._visit(swan_obj.luid, swan_obj, "luid")

    def visit_OperatorSignatureBase(
        self,
        swan_obj: swan.OperatorSignatureBase,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """OperatorSignatureBase visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Declaration(swan_obj, owner, owner_property)
        self.visit_ModuleItem(swan_obj, owner, owner_property)
        # Visit properties
        self.visit_builtin(swan_obj.is_inlined, swan_obj, "is_inlined")
        self.visit_builtin(swan_obj.is_node, swan_obj, "is_node")
        for item in swan_obj.inputs:
            self._visit(item, swan_obj, "inputs")
        for item in swan_obj.outputs:
            self._visit(item, swan_obj, "outputs")
        if swan_obj.sizes is not None:
            for item in swan_obj.sizes:
                self._visit(item, swan_obj, "sizes")
        if swan_obj.constraints is not None:
            for item in swan_obj.constraints:
                self._visit(item, swan_obj, "constraints")
        if swan_obj.specialization is not None:
            self._visit(swan_obj.specialization, swan_obj, "specialization")
        if swan_obj.pragmas is not None:
            for item in swan_obj.pragmas:
                self._visit(item, swan_obj, "pragmas")

    def visit_OptGroupItem(
        self,
        swan_obj: swan.OptGroupItem,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """OptGroupItem visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.item is not None:
            self._visit(swan_obj.item, swan_obj, "item")

    def visit_Oracle(
        self,
        swan_obj: swan.Oracle,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Oracle visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.id, swan_obj, "id")

    def visit_Partial(
        self,
        swan_obj: swan.Partial,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Partial visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_OperatorExpression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.operator, swan_obj, "operator")
        for item in swan_obj.partial_group:
            self._visit(item, swan_obj, "partial_group")

    def visit_PathIdExpr(
        self,
        swan_obj: swan.PathIdExpr,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """PathIdExpr visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.path_id, swan_obj, "path_id")

    def visit_PathIdOpCall(
        self,
        swan_obj: swan.PathIdOpCall,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """PathIdOpCall visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_OperatorBase(swan_obj, owner, owner_property)
        self.visit_PragmaBase(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.path_id, swan_obj, "path_id")

    def visit_PathIdPattern(
        self,
        swan_obj: swan.PathIdPattern,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """PathIdPattern visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Pattern(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.path_id, swan_obj, "path_id")

    def visit_PathIdentifier(
        self,
        swan_obj: swan.PathIdentifier,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """PathIdentifier visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        if isinstance(swan_obj.path_id, list):
            for item in swan_obj.path_id:
                self._visit(item, swan_obj, "path_id")
        elif SwanVisitor._is_builtin(swan_obj.path_id):
            self.visit_builtin(swan_obj.path_id, swan_obj, "path_id")

    def visit_Pattern(
        self,
        swan_obj: swan.Pattern,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Pattern visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)

    def visit_PortExpr(
        self,
        swan_obj: swan.PortExpr,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """PortExpr visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.lunum is not None:
            self._visit(swan_obj.lunum, swan_obj, "lunum")
        if swan_obj.luid is not None:
            self._visit(swan_obj.luid, swan_obj, "luid")
        if swan_obj.is_self is not None:
            self.visit_builtin(swan_obj.is_self, swan_obj, "is_self")

    def visit_Pragma(
        self,
        swan_obj: swan.Pragma,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Pragma visitor function. Should be overridden."""
        # Visit properties
        self.visit_builtin(swan_obj.pragma, swan_obj, "pragma")

    def visit_PragmaBase(
        self,
        swan_obj: swan.PragmaBase,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """PragmaBase visitor function. Should be overridden."""
        # Visit properties
        if swan_obj.pragmas is not None:
            for item in swan_obj.pragmas:
                self._visit(item, swan_obj, "pragmas")

    def visit_PredefinedType(
        self,
        swan_obj: swan.PredefinedType,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """PredefinedType visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_TypeExpression(swan_obj, owner, owner_property)

    def visit_PrefixOperatorExpression(
        self,
        swan_obj: swan.PrefixOperatorExpression,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """PrefixOperatorExpression visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_OperatorBase(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.op_expr, swan_obj, "op_expr")

    def visit_PrefixPrimitive(
        self,
        swan_obj: swan.PrefixPrimitive,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """PrefixPrimitive visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_OperatorBase(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.kind, swan_obj, "kind")

    def visit_PrefixPrimitiveKind(
        self,
        swan_obj: swan.PrefixPrimitiveKind,
        owner: Owner,
        owner_property: OwnerProperty,
    ):
        """PrefixPrimitiveKind visitor function. Should be overridden."""
        # Enum values:
        # Flatten
        # Pack
        # Reverse
        # Transpose
        pass

    def visit_ProjectionWithDefault(
        self,
        swan_obj: swan.ProjectionWithDefault,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ProjectionWithDefault visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.expr, swan_obj, "expr")
        for item in swan_obj.indices:
            self._visit(item, swan_obj, "indices")
        self._visit(swan_obj.default, swan_obj, "default")

    def visit_ProtectedDecl(
        self,
        swan_obj: swan.ProtectedDecl,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ProtectedDecl visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_ProtectedItem(swan_obj, owner, owner_property)
        self.visit_GlobalDeclaration(swan_obj, owner, owner_property)

    def visit_ProtectedExpr(
        self,
        swan_obj: swan.ProtectedExpr,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ProtectedExpr visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        self.visit_ProtectedItem(swan_obj, owner, owner_property)

    def visit_ProtectedForwardReturnItem(
        self,
        swan_obj: swan.ProtectedForwardReturnItem,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ProtectedForwardReturnItem visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_ProtectedItem(swan_obj, owner, owner_property)
        self.visit_ForwardReturnItem(swan_obj, owner, owner_property)

    def visit_ProtectedGroupRenaming(
        self,
        swan_obj: swan.ProtectedGroupRenaming,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ProtectedGroupRenaming visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_GroupRenamingBase(swan_obj, owner, owner_property)
        self.visit_ProtectedItem(swan_obj, owner, owner_property)

    def visit_ProtectedItem(
        self,
        swan_obj: swan.ProtectedItem,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ProtectedItem visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        self.visit_builtin(swan_obj.data, swan_obj, "data")
        if swan_obj.markup is not None:
            self.visit_builtin(swan_obj.markup, swan_obj, "markup")

    def visit_ProtectedOpExpr(
        self,
        swan_obj: swan.ProtectedOpExpr,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ProtectedOpExpr visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_OperatorExpression(swan_obj, owner, owner_property)
        self.visit_ProtectedItem(swan_obj, owner, owner_property)

    def visit_ProtectedPattern(
        self,
        swan_obj: swan.ProtectedPattern,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ProtectedPattern visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Pattern(swan_obj, owner, owner_property)
        self.visit_ProtectedItem(swan_obj, owner, owner_property)

    def visit_ProtectedSection(
        self,
        swan_obj: swan.ProtectedSection,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ProtectedSection visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_ScopeSection(swan_obj, owner, owner_property)
        self.visit_ProtectedItem(swan_obj, owner, owner_property)

    def visit_ProtectedTypeExpr(
        self,
        swan_obj: swan.ProtectedTypeExpr,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ProtectedTypeExpr visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_TypeExpression(swan_obj, owner, owner_property)
        self.visit_ProtectedItem(swan_obj, owner, owner_property)

    def visit_ProtectedVariable(
        self,
        swan_obj: swan.ProtectedVariable,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ProtectedVariable visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Variable(swan_obj, owner, owner_property)
        self.visit_ProtectedItem(swan_obj, owner, owner_property)

    def visit_Restart(
        self,
        swan_obj: swan.Restart,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Restart visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_OperatorExpression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.operator, swan_obj, "operator")
        self._visit(swan_obj.condition, swan_obj, "condition")

    def visit_Scope(
        self,
        swan_obj: swan.Scope,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Scope visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        self.visit_PragmaBase(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.sections is not None:
            for item in swan_obj.sections:
                self._visit(item, swan_obj, "sections")

    def visit_ScopeSection(
        self,
        swan_obj: swan.ScopeSection,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """ScopeSection visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)

    def visit_SectionBlock(
        self,
        swan_obj: swan.SectionBlock,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """SectionBlock visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_DiagramObject(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.section, swan_obj, "section")

    def visit_SensorDecl(
        self,
        swan_obj: swan.SensorDecl,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """SensorDecl visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Declaration(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.type, swan_obj, "type")

    def visit_SensorDeclarations(
        self,
        swan_obj: swan.SensorDeclarations,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """SensorDeclarations visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_GlobalDeclaration(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.sensors:
            self._visit(item, swan_obj, "sensors")

    def visit_SetSensorEquation(
        self,
        swan_obj: swan.SetSensorEquation,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """SetSensorEquation visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Equation(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.sensor, swan_obj, "sensor")
        self._visit(swan_obj.value, swan_obj, "value")

    def visit_Signature(
        self,
        swan_obj: swan.Signature,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Signature visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_OperatorSignatureBase(swan_obj, owner, owner_property)

    def visit_SizedTypeExpression(
        self,
        swan_obj: swan.SizedTypeExpression,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """SizedTypeExpression visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_TypeExpression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.size, swan_obj, "size")
        self.visit_builtin(swan_obj.is_signed, swan_obj, "is_signed")

    def visit_Slice(
        self,
        swan_obj: swan.Slice,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Slice visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.expr, swan_obj, "expr")
        self._visit(swan_obj.start, swan_obj, "start")
        self._visit(swan_obj.end, swan_obj, "end")

    def visit_Source(
        self,
        swan_obj: swan.Source,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Source visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.id, swan_obj, "id")

    def visit_State(
        self,
        swan_obj: swan.State,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """State visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_StateMachineItem(swan_obj, owner, owner_property)
        self.visit_PragmaBase(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.id is not None:
            self._visit(swan_obj.id, swan_obj, "id")
        if swan_obj.lunum is not None:
            self._visit(swan_obj.lunum, swan_obj, "lunum")
        if swan_obj.strong_transitions is not None:
            for item in swan_obj.strong_transitions:
                self._visit(item, swan_obj, "strong_transitions")
        if swan_obj.sections is not None:
            for item in swan_obj.sections:
                self._visit(item, swan_obj, "sections")
        if swan_obj.weak_transitions is not None:
            for item in swan_obj.weak_transitions:
                self._visit(item, swan_obj, "weak_transitions")
        if swan_obj.is_initial is not None:
            self.visit_builtin(swan_obj.is_initial, swan_obj, "is_initial")

    def visit_StateMachine(
        self,
        swan_obj: swan.StateMachine,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """StateMachine visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_DefByCase(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.items is not None:
            for item in swan_obj.items:
                self._visit(item, swan_obj, "items")

    def visit_StateMachineBlock(
        self,
        swan_obj: swan.StateMachineBlock,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """StateMachineBlock visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_DefByCaseBlockBase(swan_obj, owner, owner_property)

    def visit_StateMachineItem(
        self,
        swan_obj: swan.StateMachineItem,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """StateMachineItem visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)

    def visit_StateRef(
        self,
        swan_obj: swan.StateRef,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """StateRef visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.lunum is not None:
            self._visit(swan_obj.lunum, swan_obj, "lunum")
        if swan_obj.id is not None:
            self._visit(swan_obj.id, swan_obj, "id")

    def visit_StructConstructor(
        self,
        swan_obj: swan.StructConstructor,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """StructConstructor visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.group, swan_obj, "group")
        if swan_obj.type is not None:
            self._visit(swan_obj.type, swan_obj, "type")

    def visit_StructDestructor(
        self,
        swan_obj: swan.StructDestructor,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """StructDestructor visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.group, swan_obj, "group")
        self._visit(swan_obj.expr, swan_obj, "expr")

    def visit_StructField(
        self,
        swan_obj: swan.StructField,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """StructField visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.id, swan_obj, "id")
        self._visit(swan_obj.type, swan_obj, "type")

    def visit_StructProjection(
        self,
        swan_obj: swan.StructProjection,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """StructProjection visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.expr, swan_obj, "expr")
        self._visit(swan_obj.label, swan_obj, "label")

    def visit_StructTypeDefinition(
        self,
        swan_obj: swan.StructTypeDefinition,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """StructTypeDefinition visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_TypeDefinition(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.fields:
            self._visit(item, swan_obj, "fields")

    def visit_Target(
        self,
        swan_obj: swan.Target,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Target visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.target, swan_obj, "target")
        if swan_obj.is_resume is not None:
            self.visit_builtin(swan_obj.is_resume, swan_obj, "is_resume")

    def visit_TestHarness(
        self,
        swan_obj: swan.TestHarness,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """TestHarness visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Declaration(swan_obj, owner, owner_property)
        self.visit_PragmaBase(swan_obj, owner, owner_property)
        self.visit_ModuleItem(swan_obj, owner, owner_property)
        # Visit properties
        if isinstance(swan_obj.body, swan.Scope):
            self._visit(swan_obj.body, swan_obj, "body")
        elif isinstance(swan_obj.body, swan.Equation):
            self._visit(swan_obj.body, swan_obj, "body")

    def visit_TestModule(
        self,
        swan_obj: swan.TestModule,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """TestModule visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Module(swan_obj, owner, owner_property)

    def visit_Transition(
        self,
        swan_obj: swan.Transition,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Transition visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        self.visit_PragmaBase(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.arrow, swan_obj, "arrow")

    def visit_TransitionDecl(
        self,
        swan_obj: swan.TransitionDecl,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """TransitionDecl visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_StateMachineItem(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.priority is not None:
            self._visit(swan_obj.priority, swan_obj, "priority")
        self._visit(swan_obj.transition, swan_obj, "transition")
        self.visit_builtin(swan_obj.is_strong, swan_obj, "is_strong")
        self._visit(swan_obj.state_ref, swan_obj, "state_ref")

    def visit_Transpose(
        self,
        swan_obj: swan.Transpose,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Transpose visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_PrefixPrimitive(swan_obj, owner, owner_property)
        # Visit properties
        if isinstance(swan_obj.params, list):
            for item in swan_obj.params:
                self.visit_builtin(item, swan_obj, "params")
        elif SwanVisitor._is_builtin(swan_obj.params):
            self.visit_builtin(swan_obj.params, swan_obj, "params")

    def visit_TypeConstraint(
        self,
        swan_obj: swan.TypeConstraint,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """TypeConstraint visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        if isinstance(swan_obj.type_vars, list):
            for item in swan_obj.type_vars:
                self._visit(item, swan_obj, "type_vars")
        elif SwanVisitor._is_builtin(swan_obj.type_vars):
            self.visit_builtin(swan_obj.type_vars, swan_obj, "type_vars")

        self._visit(swan_obj.kind, swan_obj, "kind")

    def visit_TypeDecl(
        self,
        swan_obj: swan.TypeDecl,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """TypeDecl visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Declaration(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.definition is not None:
            self._visit(swan_obj.definition, swan_obj, "definition")

    def visit_TypeDeclarations(
        self,
        swan_obj: swan.TypeDeclarations,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """TypeDeclarations visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_GlobalDeclaration(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.types:
            self._visit(item, swan_obj, "types")

    def visit_TypeDefinition(
        self,
        swan_obj: swan.TypeDefinition,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """TypeDefinition visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)

    def visit_TypeExpression(
        self,
        swan_obj: swan.TypeExpression,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """TypeExpression visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)

    def visit_TypeGroupTypeExpression(
        self,
        swan_obj: swan.TypeGroupTypeExpression,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """TypeGroupTypeExpression visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_GroupTypeExpression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.type, swan_obj, "type")

    def visit_TypeReferenceExpression(
        self,
        swan_obj: swan.TypeReferenceExpression,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """TypeReferenceExpression visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_TypeExpression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.alias, swan_obj, "alias")

    def visit_Uint16Type(
        self,
        swan_obj: swan.Uint16Type,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Uint16Type visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_PredefinedType(swan_obj, owner, owner_property)

    def visit_Uint32Type(
        self,
        swan_obj: swan.Uint32Type,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Uint32Type visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_PredefinedType(swan_obj, owner, owner_property)

    def visit_Uint64Type(
        self,
        swan_obj: swan.Uint64Type,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Uint64Type visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_PredefinedType(swan_obj, owner, owner_property)

    def visit_Uint8Type(
        self,
        swan_obj: swan.Uint8Type,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Uint8Type visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_PredefinedType(swan_obj, owner, owner_property)

    def visit_UnaryExpr(
        self,
        swan_obj: swan.UnaryExpr,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """UnaryExpr visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.operator, swan_obj, "operator")
        self._visit(swan_obj.expr, swan_obj, "expr")

    def visit_UnaryOp(
        self,
        swan_obj: swan.UnaryOp,
        owner: Owner,
        owner_property: OwnerProperty,
    ):
        """UnaryOp visitor function. Should be overridden."""
        # Enum values:
        # Minus
        # Plus
        # Lnot
        # Not
        # Pre
        pass

    def visit_UnderscorePattern(
        self,
        swan_obj: swan.UnderscorePattern,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """UnderscorePattern visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Pattern(swan_obj, owner, owner_property)

    def visit_UseDirective(
        self,
        swan_obj: swan.UseDirective,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """UseDirective visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_ModuleItem(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.path, swan_obj, "path")
        if swan_obj.alias is not None:
            self._visit(swan_obj.alias, swan_obj, "alias")

    def visit_VarDecl(
        self,
        swan_obj: swan.VarDecl,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """VarDecl visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Declaration(swan_obj, owner, owner_property)
        self.visit_Variable(swan_obj, owner, owner_property)
        # Visit properties
        if swan_obj.is_clock is not None:
            self.visit_builtin(swan_obj.is_clock, swan_obj, "is_clock")
        if swan_obj.is_probe is not None:
            self.visit_builtin(swan_obj.is_probe, swan_obj, "is_probe")
        if swan_obj.type is not None:
            self._visit(swan_obj.type, swan_obj, "type")
        if swan_obj.when is not None:
            self._visit(swan_obj.when, swan_obj, "when")
        if swan_obj.default is not None:
            self._visit(swan_obj.default, swan_obj, "default")
        if swan_obj.last is not None:
            self._visit(swan_obj.last, swan_obj, "last")

    def visit_VarSection(
        self,
        swan_obj: swan.VarSection,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """VarSection visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_ScopeSection(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.var_decls:
            self._visit(item, swan_obj, "var_decls")

    def visit_Variable(
        self,
        swan_obj: swan.Variable,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Variable visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)

    def visit_VariableTypeExpression(
        self,
        swan_obj: swan.VariableTypeExpression,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """VariableTypeExpression visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_TypeExpression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.name, swan_obj, "name")

    def visit_VariantComponent(
        self,
        swan_obj: swan.VariantComponent,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """VariantComponent visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_SwanItem(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.tag, swan_obj, "tag")

    def visit_VariantPattern(
        self,
        swan_obj: swan.VariantPattern,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """VariantPattern visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Pattern(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.path_id, swan_obj, "path_id")
        if swan_obj.captured is not None:
            self._visit(swan_obj.captured, swan_obj, "captured")
        if swan_obj.underscore is not None:
            self.visit_builtin(swan_obj.underscore, swan_obj, "underscore")

    def visit_VariantSimple(
        self,
        swan_obj: swan.VariantSimple,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """VariantSimple visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_VariantComponent(swan_obj, owner, owner_property)

    def visit_VariantStruct(
        self,
        swan_obj: swan.VariantStruct,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """VariantStruct visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_VariantComponent(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.fields:
            self._visit(item, swan_obj, "fields")

    def visit_VariantTypeDefinition(
        self,
        swan_obj: swan.VariantTypeDefinition,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """VariantTypeDefinition visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_TypeDefinition(swan_obj, owner, owner_property)
        # Visit properties
        for item in swan_obj.tags:
            self._visit(item, swan_obj, "tags")

    def visit_VariantTypeExpr(
        self,
        swan_obj: swan.VariantTypeExpr,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """VariantTypeExpr visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_VariantComponent(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.type, swan_obj, "type")

    def visit_VariantValue(
        self,
        swan_obj: swan.VariantValue,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """VariantValue visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.tag, swan_obj, "tag")
        self._visit(swan_obj.group, swan_obj, "group")

    def visit_WhenClockExpr(
        self,
        swan_obj: swan.WhenClockExpr,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """WhenClockExpr visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.expr, swan_obj, "expr")
        self._visit(swan_obj.clock, swan_obj, "clock")

    def visit_WhenMatchExpr(
        self,
        swan_obj: swan.WhenMatchExpr,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """WhenMatchExpr visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.expr, swan_obj, "expr")
        self._visit(swan_obj.when, swan_obj, "when")

    def visit_Window(
        self,
        swan_obj: swan.Window,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Window visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_Expression(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.size, swan_obj, "size")
        self._visit(swan_obj.init, swan_obj, "init")
        self._visit(swan_obj.params, swan_obj, "params")

    def visit_Wire(
        self,
        swan_obj: swan.Wire,
        owner: Owner,
        owner_property: OwnerProperty,
    ) -> None:
        """Wire visitor function. Should be overridden."""
        # Visit base class(es)
        self.visit_DiagramObject(swan_obj, owner, owner_property)
        # Visit properties
        self._visit(swan_obj.source, swan_obj, "source")
        for item in swan_obj.targets:
            self._visit(item, swan_obj, "targets")
